type level = int
type tycon = string

datatype typeterm =
    TVar of level
  | TCons of tycon * ty list
  | TPtr of ty
withtype ty = typeterm ref

exception Unification of string   (* error message *)

val level = ref 0

fun var () = ref (TVar (!level))

fun cons (tycon, tys) = ref (TCons (tycon, tys))

fun unify (ty, ty') = ...

fun instantiate ty =
    let
	val env = ref nil
	fun instantiate' (r as ref (TVar ~1)) =
	    (case List.find (fn (r', _) => r = r') (!env) of
		 SOME (_, r') => r'
	       | NONE =>
		     let
			 val r' = var ()
		     in
			 env := (r, r')::(!env); r'
		     end)
	  | instantiate' (r as ref (TVar _)) = r
	  | instantiate' (ref (TCons (tycon, tys))) =
	    cons (tycon, List.map instantiate' tys)
	  | instantiate' (ref (TPtr ty)) = instantiate' ty
    in
	instantiate' ty
    end

fun beginScheme () = level := !level + 1

fun endScheme ty =
    let
	val n = !level
	fun generalize (r as ref typeterm) =
	    case typeterm of
		TVar i => if i = n then r := TVar ~1 else ()
	      | TCons (_, tys) => List.app generalize tys
	      | TPtr ty => generalize ty
    in
	level := n - 1;
	generalize ty
    end

local
    exception Crash

    fun member (x, xs) = List.exists (fn y => y = x) xs

    fun adjoin (x, xs) = if member (x, xs) then xs else x::xs

    fun union (xs, ys) = List.foldr adjoin xs ys

    fun collect (r as ref (TVar _), rs) = adjoin (r, rs)
      | collect (r as ref (TCons (_, tys)), rs) =
	List.foldl collect rs tys
      | collect (r as ref (TPtr ty), rs) =
	collect (ty, rs)
in
    fun toString ty =
	let
	    val counter = ref (Char.ord #"a")
	    val vars =
		List.foldr (fn (r, rs) =>
			    let
				val n = !counter
			    in
				counter := n + 1;
				(r, "'" ^ String.str (Char.chr n))::rs
			    end) nil (collect (ty, nil))
	    val quantified =
		List.foldr (fn ((ref (TVar ~1), s), s') => " " ^ s ^ s'
			     | ((ref _, _), s') => s') "" vars
	    fun toString' (r as ref (TVar _), _) =
		(case List.find (fn (r', _) => r' = r) vars of
		     SOME (_, s) => s
		   | NONE => raise Crash)
	      | toString' (r as ref (TCons ("->", [ty1, ty2])), true) =
		"(" ^ toString' (r, false) ^ ")"
	      | toString' (ref (TCons ("->", [ty1, ty2])), false) =
		toString' (ty1, true) ^ " -> " ^ toString' (ty2, false)
	      | toString' (ref (TCons (tycon, nil)), _) = tycon
	      | toString' (ref (TCons (tycon, ty::tys)), _) =
		tycon ^ "(" ^ (List.foldl
			       (fn (ty, s) => s ^ ", " ^ toString' (ty, false))
			       (toString' (ty, false)) tys) ^ ")"
	      | toString' (ref (TPtr ty), b) = toString' (ty, b)
	in
	    case quantified of
		"" => toString' (ty, false)
	      | s => "forall" ^ s ^ ": " ^ toString' (ty, false)
	end
end
